from typing import Any


def construct_confident_event(
    payload: Any, confident_ai_signature: str, confident_secret: str
):
    raise ValueError("Please contact support@confident-ai.com for beta access.")
